/*
 * This is an example of the Tcl/C++ interface.
 *
 * We define a C++ class "fraction", a numeric data type to represent
 * fractional numbers, with the numerator and denominator both a long.
 *
 * On the Tcl level, we implement the functions "add", "sub", "div"
 * and "mult", using the class's operators
 *
 * If HAVE_TCLOBJ is not defined, this class will just be a C++ class
 * like any other.
 */

#ifdef HAVE_TCLOBJ
#include <tcl.h>
#include <tclobj.h>
#endif

#include <stdio.h>
#include "fraction.h"

/*
 * euklid's algorithm. also adjust the sign.
 */

fraction &
fraction::gcf (void)
{
  long sign=1, mod, first, second;

  if (den<0) { num=-num; den=-den; }

  first  = (num<=0)?(-num):num;
  second = (den==0)?1:den;

  while (second) {
    mod    = first % second;
    first  = second;
    second = mod;
  }

  /*
   * gcf is in first
   */

  num /= first;
  den /= first;

  return *this;
}

/*
 * The other arithmetics. I agree the design isn't perfect.
 */

fraction &
fraction::operator+= (const fraction &other)
{
  long newden = den * other.den;

  num = num*other.den + den*other.num;
  den = newden;

  return gcf();
}

fraction &
fraction::operator*= (const fraction &other)
{
  num *= other.num;
  den *= other.den;

  return gcf();
}

int
fraction::operator< (const fraction &other) const
{
  return (operator double() < (double)other);
}

fraction::operator double() const
{
  return ((double)num/(double)den);
}

fraction::operator const char*() const
{
  static char buf[32];
  sprintf (buf, "%f", operator double());
  return buf;
}

/*
 * That was easy, wasn't it? Now define the Tcl interface
 */

#ifdef HAVE_TCLOBJ

/*
 * An initialization procedure
 */

extern "C" int
Fraction_Init (Tcl_Interp *interp)
{
  fraction * oneseventh = new fraction (1,7);
  oneseventh->Tcl_Enable (interp, "oneseventh");
  return TCL_OK;
}

/*
 * our constructors
 */

static const TclObj_Comtab fctab[] = {
  {  0, NULL, NULL, NULL }
};

/*
 * our functions
 */

static const TclObj_Comtab fftab[] = {
  /* constructors */
  {  1, "new", "",         "create fraction of value zero" },
  {  2, "new", "int",      "create fraction of integer value" },
  {  3, "new", "int int",  "create fraction of numerator and denominator" },
  {  4, "new", "fraction", "copy constructor" },
  /* operators */
  {  1, "+=", "fraction",  "add fractional value" },
  {  2, "-=", "fraction",  "substract fractional value" },
  {  3, "*=", "fraction",  "multiply with fractional value" },
  {  4, "/=", "fraction",  "divide by fractional value" },
  {  5, "<",  "fraction",  "true if value is less than other" },
  {  6, "<=", "fraction",  "true if value is less or equal than other" },
  {  7, "==", "fraction",  "true if value is equal to other" },
  {  8, "!=", "fraction",  "true if value is not equal to other" },
  {  9, ">=", "fraction",  "true if value is greater or equal than other" },
  { 10, ">",  "fraction",  "true if value is greater than other" },
  { 11, "=",  "fraction",  "set to fractional value" },
  { 12, "=",  "int",       "set to integer value" },
  { 13, "=",  "int int",   "set to numerator and denominator" },
  { 14, "value", "",       "print value as floating-point number" },
  { 15, "numden", "",      "get pair of numerator and denominator" },
  { 0, NULL, NULL, NULL }
};

/*
 * Declare ourselves
 */

TCL_OBJECT_DECL_BODY(fraction,Tcl_Object,fftab,Fraction_Init);

/*
 * Tcl-Style constructor
 */

fraction::fraction (Tcl_Interp *interp, int argc, char *argv[], int command)
{
  const fraction *other;

  switch (command) {
  case 1: /* new */
    num = 0;
    den = 1;
    break;

  case 2: /* new int */
    num = TCL_OBJECT_ARG_INT(interp, argv[2]);
    den = 1;
    break;

  case 3: /* new int int */
    num = TCL_OBJECT_ARG_INT(interp, argv[2]);
    den = TCL_OBJECT_ARG_INT(interp, argv[3]);
    break;

  case 4: /* new fraction */
    other = TCL_OBJECT_ARG_OBJECT(interp, argv[2], fraction);
    num   = other->num;
    den   = other->den;
    break;
  }
}

/*
 * The interface procedure
 */

int
fraction::Tcl_Proc (Tcl_Interp *interp,
		    int argc, char *argv[],
		    int command)
{
  const fraction *other;
  char tmpstr[32];

  /*
   * Commands 1 .. 11 all have a fraction object as parameter
   */

  if (command < 12) {
    other = TCL_OBJECT_ARG_OBJECT(interp, argv[2], fraction);
  }

  switch (command) {
  case 1: /* add fraction */
    return TCL_OBJECT_RETURN_OBJECT(interp, &operator+=(*other));

  case 2: /* sub fraction */
    return TCL_OBJECT_RETURN_OBJECT(interp, &operator-=(*other));

  case 3: /* mul fraction */
    return TCL_OBJECT_RETURN_OBJECT(interp, &operator*=(*other));

  case 4: /* div fraction */
    return TCL_OBJECT_RETURN_OBJECT(interp, &operator/=(*other));

  case 5: /* lt fraction */
    return TCL_OBJECT_RETURN_INT(interp, operator< (*other));

  case 6: /* le fraction */
    return TCL_OBJECT_RETURN_INT(interp, operator<=(*other));

  case 7: /* eq fraction */
    return TCL_OBJECT_RETURN_INT(interp, operator==(*other));

  case 8: /* ne fraction */
    return TCL_OBJECT_RETURN_INT(interp, operator!=(*other));

  case 9: /* ge fraction */
    return TCL_OBJECT_RETURN_INT(interp, operator>=(*other));

  case 10: /* gt fraction */
    return TCL_OBJECT_RETURN_INT(interp, operator> (*other));

  case 11: /* = fraction */
    return TCL_OBJECT_RETURN_OBJECT(interp, &operator=(*other));

  case 12: /* = int */
    num = TCL_OBJECT_ARG_INT(interp, argv[2]);
    den = 1;
    return TCL_OBJECT_RETURN_OBJECT(interp, this);

  case 13: /* = int int */
    num = TCL_OBJECT_ARG_INT(interp, argv[2]);
    den = TCL_OBJECT_ARG_INT(interp, argv[3]);
    return TCL_OBJECT_RETURN_OBJECT(interp, &gcf());

  case 14: /* value */
    return TCL_OBJECT_RETURN_DOUBLE(interp, operator double());

  case 15: /* numden */
    return TclObj_MkResult (interp, "long long", num, den);
  }
  /*
   * emergency exit
   */
  TCL_OBJECT_NO_MATCH(interp);
}

#endif
