/*
 * Maintain registry of classes. The state is stored in the
 * interpreter (as Tcl variables), so we can register classes
 * in each interpreter separately.
 */

#include <tcl.h>
#include <stdlib.h>
#include "tclobj.h"
#include "register.h"
#include "compat.h"
#include "signature.h"

/*
 * cvs version
 */
char * tclobj_register_id = "$Id: register.cc,v 1.3 1997/01/09 16:18:39 fp Exp $";


/*
 * next available identifier. we cannot store this in the interpreter,
 * because we expect the identifiers to be unique over interpreter
 * boundaries.
 */

static unsigned long TclObj_NextFreeIdentifier = 256;

/*
 * reserve a new identifier number for a not-yet-available class.
 * this is necessary for classes that depend on one another.
 * In this case, we set TclObj_ClassInfo(name) to 0 and set
 * TclObj_Unregistered(name) to the identifier.
 */

unsigned long
TclObj_RegisterIdentifier (Tcl_Interp *interp, const char *name)
{
  char tmpstr[32], *ptrname;
  int theidentifier;

  if ((ptrname = Tcl_GetVar2 (interp, "TclObj_ClassInfo", 
			      (char *) name,
			      TCL_GLOBAL_ONLY)) == NULL) {
    /*
     * does not yet exist
     */

    theidentifier = TclObj_NextFreeIdentifier++;

    sprintf (tmpstr, "%lu", (unsigned long) theidentifier);
    Tcl_SetVar2 (interp, "TclObj_ClassInfo", 
		 (char *) name, "0",
		 TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "TclObj_Unregistered",
		 (char *) name, tmpstr,
		 TCL_GLOBAL_ONLY);

    return theidentifier;
  }
  
  /*
   * does exist
   */

  if (strtoul (ptrname, NULL, 0) == 0) {
    /*
     * but is not yet registered
     */
    if ((ptrname = Tcl_GetVar2 (interp, "TclObj_Unregistered",
				(char *) name,
				TCL_GLOBAL_ONLY)) == NULL) {
      /* oops, shouldn't be here */
      return 0;
    }

    return strtoul (ptrname, NULL, 0);
  }

  /*
   * hey, but it is already registered
   */

  return ((TclObj_Info*) strtoul (ptrname, NULL, 0))->ident;
}

/*
 * register new class with the interpreter, initializing it if
 * necessary.
 */

unsigned long
TclObj_RegisterClass (Tcl_Interp *interp, TclObj_Info *theinfo)
{
  char tmpstr[32], str2[32], *ptr;
  unsigned long regcount;
  int needtoinitialize=0;

  if (theinfo->ident == 0) {
    /*
     * see if it was expected. If not, assign new identifier
     */
    if ((ptr = Tcl_GetVar2 (interp, "TclObj_Unregistered",
			    (char *) theinfo->name,
			    TCL_GLOBAL_ONLY)) != NULL) {
      theinfo->ident = strtoul (ptr, NULL, 0);
      Tcl_UnsetVar2 (interp, "TclObj_Unregistered",
		     (char *) theinfo->name,
		     TCL_GLOBAL_ONLY);
    }
    else {
      theinfo->ident   = TclObj_NextFreeIdentifier++;
    }
    theinfo->counter = 0;
    needtoinitialize = 1;
  }

  sprintf     (tmpstr, "%lu", (unsigned long) theinfo);
  Tcl_SetVar2 (interp, "TclObj_ClassInfo", (char *) theinfo->name,
	       tmpstr, TCL_GLOBAL_ONLY);

  sprintf     (str2,   "%lu", (unsigned long) theinfo->ident);
  Tcl_SetVar2 (interp, "TclObj_ClassInfo", str2, tmpstr,
	       TCL_GLOBAL_ONLY);

  if (needtoinitialize) {
    if (TclObj_BuildComptree (interp, theinfo) != TCL_OK) {
      Tcl_AppendResult (interp, "registration of class \"",
			(char *) theinfo->name, "\" failed.",
			NULL);
      return 0;
    }
  }

  return theinfo->ident;
}

/*
 * Query info
 */

TclObj_Info *
TclObj_GetClassInfo (Tcl_Interp *interp, unsigned long ident)
{
  char *ptrname, tmpstr[32];
  
  sprintf (tmpstr, "%lu", (unsigned long) ident);
  if ((ptrname = Tcl_GetVar2 (interp, "TclObj_ClassInfo", tmpstr,
			      TCL_GLOBAL_ONLY)) == NULL)
    return NULL;
  
  return (TclObj_Info*) strtoul (ptrname, NULL, 0);
}

TclObj_Info *
TclObj_GetClassInfo (Tcl_Interp *interp, const char *thetype)
{
  char *ptrname;
  
  if ((ptrname = Tcl_GetVar2 (interp, "TclObj_ClassInfo", 
			      (char *) thetype,
			      TCL_GLOBAL_ONLY)) == NULL)
    return NULL;

  return (TclObj_Info*) strtoul (ptrname, NULL, 0);
}

unsigned long
TclObj_GetIdentifier (Tcl_Interp *interp, const char *thetype)
{
  TclObj_Info *theinfo;
  char *ptr;

  if ((theinfo = TclObj_GetClassInfo (interp, thetype)) != NULL)
    return theinfo->ident;

  /*
   * see if it was expected
   */

  if ((ptr = Tcl_GetVar2 (interp, "TclObj_Unregistered",
			  (char *) thetype,
			  TCL_GLOBAL_ONLY)) == NULL) {
    /* oops, shouldn't be here */
    return 0;
  }

  return strtoul (ptr, NULL, 0);
}
